/*---------------------------------------------------------------------------
  #######      #     #########   #          #   #    ########     #    #    #
  #            #     #            #        #    #    #            #    #    #
   ###         #    #              #      #     #    #####         #  # #  #
      ###      #    #    #####     #     #      #    #             #  # #  #
         #     #    #        #      #   #       #    #              # #  # #
  ##     #     #     ##      #       # #        #    #              ##   ##
    ####       #       ######         #         #    ########        #    #

    Designed and written by
    Lisa Pickel - pickel@mcnc.org 

    Copyright 1990, 1991 by MCNC.
    This program may be freely distributed provided that the
	copyright notice and author identification are retained.
--------------------------------------------------------------------------- */
#include "sigview.h"

comm_file_t *Sig_files;
char **Sig_signms;
int Sig_signm_ct;
char **GP_vecnames;
int GP_vecname_ct;

extern axis_t *Sig_axes;
extern vert_t Sig_leftvert, Sig_rightvert;
extern int GP_mode;

/* Returns 0 if we couldn't load or reload the file, else return 1 */
int
load_simfile(fname)
	char *fname;
{
	FILE *simfile, *open_fl();
	int type, success;
	comm_file_t *file_ptr, *allocate_space();

	sig_file_msg(0, "Sigview about to load ",fname);

	if (!(simfile = open_fl(fname, "r"))) {
		sig_file_msg(1, "Load: can't find or open file ",fname);
		return(0);
	}	

	if (!(type = which_type(simfile, fname))) {
		sig_file_msg(1, "Load: cannot identify type of ",fname);
		return(0);
	}
	if (!(file_ptr = allocate_space()))
		return(0);

	file_ptr->name[0] = '\0';
	assign_filename(fname, file_ptr);
	file_ptr->type = type;

	switch(type) {
		case CAZM:
			sig_file_msg(0, "Assumed file to be CAzM format","");
			success = load_analog(simfile, file_ptr);
			break;
		case SPICE:
			sig_file_msg(0, "Assumed file to be SPICE format","");
			success = load_analog(simfile, file_ptr);
			break;
		case RAW:
			sig_file_msg(0, "Assumed file to be CAzM, SPICE or Spectre RAW format","");
			success = load_raw(simfile, file_ptr);
			break;
		case RNL:
			sig_file_msg(0, "Assumed file to be RNL format","");
			success = load_rnl(simfile, file_ptr);
			break;
		case HILO:
			sig_file_msg(0, "Assumed file to be HILO format","");
			success = load_hilo(simfile, file_ptr);
			break;
		case ESIM:
			sig_file_msg(0, "Assumed file to be ESIM format","");
			success = load_esim(simfile, file_ptr);
			break;
		case HSPICE:
			sig_file_msg(0, "Assumed file to be HSPICE format","");
			success = load_analog(simfile, file_ptr);
			break;
		default:
			sig_file_msg(1, "Load: could not identify type of ",fname);
			success = 0;
	}
	fclose(simfile);

	if (!success) 
		deallocate_space(fname);

	if (Sig_leftvert.run == NULL) {
		Sig_leftvert.run = file_ptr->runs;
		Sig_rightvert.run = file_ptr->runs;
	}
/* dump(file_ptr);   */
	
	add_global_signms(file_ptr);
	
	return(success);
}


/* Look at first line of simfile (already opened).  Identify it's type
   based on what we find in this line.  If we can't identify it from that,
   search through for the word "HILO" to see if it is a Hilo file.
   Else, we can't identify it and we return 0.
*/
static int
which_type(simfile, filename)
	FILE *simfile;
	char *filename;
{
	char line[BUFSIZE];

	rewind(simfile); /* insure we are at top line */

	if (fgets(line, BUFSIZE, simfile) == NULL) 
		return(0);

	rewind(simfile);
	
/* Changed order of evaluation for filetype -- JTS AWG */

	if (strncmp(line, "Title:", 6) == 0) {
		return(RAW);
	}
	else if (strncmp(line, ">>>", 3) == 0) {
		return(RNL);
	}
	else if ((substr(line, "CAzM", 0, strlen(line)) != -1) ||
		(substr(line, "CAZM", 0, strlen(line)) != -1) ||
		(strncmp(line, "Time", 4) == 0) ||
		(strncmp(line, "AC ANALYSIS", 11) == 0) ||
		(strncmp(line, "TRANSFER", 8) == 0) ||
		(strncmp(line, "TRANSIENT", 9) == 0) ||
		(strncmp(line, "DC ANALYSIS", 11) == 0)) {
		return(CAZM);
	}
    else if (substr(line, "H S P I C E", 0, strlen(line)) != -1) {
        return(HSPICE);
    }
	else if (substr(line, "SPICE", 0, strlen(line)) != -1) {
		return(SPICE);
	}
	else if ((substr(line, "transistors, ", 0, strlen(line)) != -1) ||
			 (substr(line, "step took", 0, strlen(line)) != -1)) {
		return(ESIM);
	}
	else if (its_hilo(simfile)) {
		return(HILO);
	}
	/* add this case in case people have the old version of CAzM which
	   has no cues in-file like "TRANSFER" or "TRANSIENT" etc.  It 
	   expects that the filename will have the extension ".cta" which
	   means we have a transfer analysis file here. */
	else if (substr(filename, ".cta", 0, strlen(filename)) != -1) {
		return(CAZM);
	}
	else if (substr(filename, ".post", 0, strlen(filename)) != -1) {
		return(CAZM);
	}
	else {
		return(0);
	}
}

static int
its_hilo(simfile)
	FILE *simfile;
{
	char line[BUFSIZE];

	while (fgets(line, BUFSIZE, simfile) != NULL) {
		if (strncmp(line, "HILO ", 5) == 0) {
			rewind(simfile);
			return(1);
		}
	}
	rewind(simfile);
	return(0);
}

comm_file_t *
allocate_space()
{
	comm_file_t *file_ptr, *previous;

	if (Sig_files == NULL) {
		if ((Sig_files = (comm_file_t*) malloc(sizeof(comm_file_t))) == NULL)
			return(NULL);
		Sig_files->next = NULL;
		Sig_files->runs = NULL;
		file_ptr = Sig_files;
	}

	else {
		file_ptr = Sig_files;
		while (file_ptr != NULL) {
			previous = file_ptr;
			file_ptr = file_ptr->next;
		}

		file_ptr = Sig_files;
		while (file_ptr->next != NULL)
			file_ptr = file_ptr->next;
		if ((file_ptr = (comm_file_t*) malloc(sizeof(comm_file_t))) == NULL)
			return(NULL);
		file_ptr->next = NULL;
		file_ptr->runs = NULL;
		previous->next = file_ptr;
	}
	return(file_ptr);
}

/* Here we assume that we have already trashed the X, Y, and Ynames
   arrays in the routines previous. */

static
deallocate_space(filename)
	char *filename;
{
	comm_file_t *file_ptr, *previous;

	/* since we put this file structure here, we are guaranteed to find it. */
	file_ptr = Sig_files;
	previous = NULL;

	while ((file_ptr != NULL) && (strcmp(file_ptr->name, filename) != 0)) {
		previous = file_ptr;
		file_ptr = file_ptr->next;
	}

	if (previous != NULL) 
		previous->next = file_ptr->next;
	else
		Sig_files = file_ptr->next;

	deallocate_guts(file_ptr);
	free((char *)file_ptr);
}

/* called when something went wrong during file read-in.  Check to see 
   if there is anything allocated for the arrays and deallocate them. */

static
deallocate_guts(file_ptr)
	comm_file_t *file_ptr;
{
	run_t *run_ptr;

	run_ptr = file_ptr->runs;
	while (run_ptr != NULL)  {
		deallocate_guts2(run_ptr);
		run_ptr = run_ptr->next;
	}
}

deallocate_guts2(run_ptr)
	run_t *run_ptr;
{
	int i;
	dig_run_t *dig_ptr;
	double ** anal_ptr;

	if (run_ptr->Ynames) {
		for (i = 0; i < run_ptr->num_Y; i++) {
			if (run_ptr->Ynames[i])
				free((char *)run_ptr->Ynames[i]);
		}
		if (run_ptr->num_Y > 0)
			free((char *)run_ptr->Ynames);
	}	
	if (run_ptr->run_type == DIGITL) {
		dig_ptr = &(run_ptr->data.digital);
		if (dig_ptr->X)
			free((char *)dig_ptr->X);
		if (dig_ptr->Y) {
			for (i = 0; i < run_ptr->num_Y; i++) {
				if (dig_ptr->Y[i])
					free((char *)dig_ptr->Y[i]);
			}
			free((char *)dig_ptr->Y);
		}
	}
	else {
		anal_ptr = run_ptr->data.analog;
		if (anal_ptr) {
			for (i = 0; i < run_ptr->num_Y; i++) {
				if (anal_ptr[i])
					free((char *)anal_ptr[i]);
			}	
			free((char *)anal_ptr);
		}
	}
}

assign_run_layer(run_ptr)
	run_t *run_ptr;
{
	static next_layer = 4;

	/* if this file is being reloaded, use the same layer it already 
	   has.  Else assign a new one (structure has -1 for layer if never 
	   loaded before). */

	if (run_ptr->layer == -1) {
		run_ptr->layer = next_layer;
		if (next_layer == 9)
			next_layer = 4;
		else next_layer++;
	}
}

assign_runname(run_ptr, name, runnumber)
	run_t *run_ptr;
	char *name;
	int runnumber;
{
	char holder[SIGNMSIZE];

	if (runnumber == 0)
		strcpy(run_ptr->name, name);
	else {
		sprintf(holder, "%s%c%d", name, UNIQUECHAR, runnumber);
		strcpy(run_ptr->name, holder);
	}
}

assign_filename(filename, fileptr)
	char *filename;
	comm_file_t *fileptr;
{
	int dup_ct;
	comm_file_t *file_p;
	char existing_nm[SIGNMSIZE];

	strip_subscript(filename);

	/* count the number of files which have this same name */
	dup_ct = 0;
	file_p = Sig_files;
	while (file_p != NULL) {
		if (file_p->name[0] != '\0') {
			strcpy(existing_nm, file_p->name);
			/* first, strip name of possible subscript */
			strip_subscript(existing_nm);

			if (strcmp(existing_nm, filename) == 0)
				dup_ct++;
		}
		file_p = file_p->next;
	}

	if (dup_ct == 0)
		strcpy(fileptr->name, filename);
	else sprintf(fileptr->name, "%s%c%d", filename, UNIQUECHAR, dup_ct);
}

dump(file_ptr)
	comm_file_t *file_ptr;
{
	int i,j;
	dig_run_t *data;
	double **dataa;
	run_t *run_ptr;

	fprintf(stderr, "\tFile: %s\n", file_ptr->name);
	fprintf(stderr, "\t\tFile type: %d\n", file_ptr->type);

	run_ptr = file_ptr->runs;
	while (run_ptr != NULL) {
	
		if (run_ptr->run_type == DIGITL) 
			data = &(run_ptr->data.digital);
		else dataa = run_ptr->data.analog;

		fprintf(stderr, "\tRunName: %s\n", run_ptr->name);
		fprintf(stderr, "\t\tNumber of X: %d\n", run_ptr->num_X);
		fprintf(stderr, "\t\tNumber of Y: %d\n", run_ptr->num_Y);
		fprintf(stderr, "\t\tRun type: %d\n", run_ptr->run_type);
		fprintf(stderr, "\t\tLayer: %d\n\n", run_ptr->layer);

		for (j = 0; j < run_ptr->num_Y; j++)
			fprintf(stderr, "\t%s", run_ptr->Ynames[j]);
		fprintf(stderr, "\n");
	
		if (run_ptr->run_type == DIGITL) {
			for (i=0; i < run_ptr->num_X; i++) {
				fprintf(stderr, "\t%e\t", data->X[i]);
				for (j = 0; j < run_ptr->num_Y; j++)
					fprintf(stderr, "%c\t", data->Y[j][i]);
				fprintf(stderr, "\n");
			}
		}
		else {
			for (i=0; i < run_ptr->num_X; i++) {
				for (j = 0; j < run_ptr->num_Y; j++)
					fprintf(stderr, "%e\t", dataa[j][i]);
				fprintf(stderr, "\n");
			}
		}
		fprintf(stderr, "\n");
		run_ptr = run_ptr->next;
	}
}


add_global_signms(file_ptr)
	comm_file_t *file_ptr;
{
	run_t *run_ptr;
	int i, j, already_there;

	run_ptr = file_ptr->runs;
	while (run_ptr != NULL) {
		if (run_ptr->run_type != DIGITL)
			add_gp_vecnames(run_ptr);

		if (Sig_signms == NULL)
			Sig_signms = (char **) calloc(run_ptr->num_Y, sizeof(char *));
		else Sig_signms = (char **) realloc((char *)Sig_signms,
			(Sig_signm_ct + run_ptr->num_Y) * sizeof(char *));

		/* for each Yname in run_ptr, see if it is already in Sig_signms.
		   If not, add it to list and increment Sig_signm_ct */
		if (run_ptr->run_type != DIGITL)
			i = 1; /* to skip the name of the first col. e.g. "Time" */
		else i= 0;

		while (i < run_ptr->num_Y) {
			already_there = 0;
			j = 0;
			while ((j < Sig_signm_ct) && (!already_there)) {
				if (strcmp(run_ptr->Ynames[i], Sig_signms[j]) == 0)
					already_there = 1;
				j++;
			}
			if (!already_there) {
				Sig_signms[Sig_signm_ct] = (char *)
					calloc(strlen(run_ptr->Ynames[i]) + 1, sizeof(char));
				strcpy(Sig_signms[Sig_signm_ct], run_ptr->Ynames[i]);
				Sig_signm_ct++;
			}
			i++;
		}
		Sig_signms = (char **) realloc((char *)Sig_signms,
			Sig_signm_ct* sizeof(char *));
		run_ptr = run_ptr->next;
	}
}

add_gp_vecnames(run)
	run_t *run;
{
	int i, j, already_there;

	if (GP_vecnames == NULL)
		GP_vecnames = (char **) calloc(run->num_Y, sizeof(char *));
	else GP_vecnames = (char **) realloc((char *)GP_vecnames,
		(GP_vecname_ct + run->num_Y) * sizeof(char *));

	for (i = 0; i < run->num_Y; i++) {
		already_there = 0;
		j = 0;
		while ((j < GP_vecname_ct) && (!already_there)) {
			if (strcmp(run->Ynames[i], GP_vecnames[j]) == 0)
				already_there = 1;
			j++;
		}
		if (!already_there) {
			GP_vecnames[GP_vecname_ct] = (char *)
				calloc(strlen(run->Ynames[i]) + 1, sizeof(char));
			strcpy(GP_vecnames[GP_vecname_ct], run->Ynames[i]);
			GP_vecname_ct++;
		}
	}
	GP_vecnames = (char **) realloc((char *)GP_vecnames, 
			GP_vecname_ct * sizeof(char *));
}

run_t *
which_run(runname)
	char *runname;
{
	comm_file_t *file_ptr;
	run_t *run_ptr;

	run_ptr = NULL;

	file_ptr = Sig_files;
	while ((file_ptr != NULL) && (run_ptr == NULL)) {
		run_ptr = file_ptr->runs;
		while ((run_ptr != NULL)  && (strcmp(run_ptr->name, runname) != 0))
			run_ptr = run_ptr->next;
		file_ptr = file_ptr->next;
	}
	return(run_ptr);
}

run_t *
last_run()
{
	comm_file_t *file_ptr;
	run_t *run_ptr;

	if (Sig_files == NULL)
		return(NULL);

	file_ptr = Sig_files;
	while (file_ptr->next != NULL) 
		file_ptr = file_ptr->next;

	run_ptr = file_ptr->runs;
	if (run_ptr == NULL)
		return(NULL);

	while (run_ptr->next != NULL) 
		run_ptr = run_ptr->next;
	return(run_ptr);
}
