/*---------------------------------------------------------------------------
  #######      #     #########   #          #   #    ########     #    #    #
  #            #     #            #        #    #    #            #    #    #
   ###         #    #              #      #     #    #####         #  # #  #
      ###      #    #    #####     #     #      #    #             #  # #  #
         #     #    #        #      #   #       #    #              # #  # #
  ##     #     #     ##      #       # #        #    #              ##   ##
    ####       #       ######         #         #    ########        #    #

    Designed and written by
    Lisa Pickel - pickel@mcnc.org 

    Copyright 1990, 1991 by MCNC.
    This program may be freely distributed provided that the
	copyright notice and author identification are retained.
--------------------------------------------------------------------------- */

#include "sigview.h"
#include <ctype.h>

/* ASSUMED SPICE FILE FORMAT:
   0) The word SPICE (all caps) must be somewhere in the first line of
	  the file.

   1) Multiple runs may be present in one file.  Each run is delimited by
	  one of the following section headers (beginning in the first column):

	  0****       DC TRANSFER CURVES  ........
	  0****       TRANSIENT ANALYSIS ..........
	  0****       AC ANALYSIS ..........
						
						OR by
	  ****       DC TRANSFER CURVES  ........
	  ****       TRANSIENT ANALYSIS ..........
	  ****       AC ANALYSIS ..........

	   Namely, a star is searched for in the second column.
	   Lines beginning this way are searched for the run type strings such
	   as AC ANALYSIS or whatever.	
	   
	2) A section headed by one of the above lines which is, in actuality,
	   one of those plots of data and not the data itself will cause royal
	   rumpus.  These plotting segments should NOT be present in your file.

	3) A section header ends with a line of stars.

	4) The line of stars is followed by optional blank lines and then
	   the signal names.  All the signal names will be on one line.  
	   NOTE: if the names contain any uppercase letters, they will be 
	   transformed to lowercase.  This,
	   as guaranteed by Gary Nifong, will be okay and is done to allow
	   easy "compare"ing with CAZM signal names.

	5) The signal name line is followed by optional blank lines and then
	   the data.  The data is read until a non-numeric field is read.
*/

/* ASSUMED CAZM FILE FORMAT:

   From CAZM version 4.0:
   1) Only one run may be present in one file.

   2) The signal names must begin on the first line of the file.

   3) A transient analysis file can have any name.  A transfer 
	  analysis file name must have the extension ".cta".


   From CAZM version 4.1:
   1) Multiple runs may be present in one file.  Each run is delimited by
	  one of the following section headers (beginning in the first column):

	  "Time" (which identifies transient analysis section)
	  "TIME"
	  "TRANSIENT"
	  "AC ANALYSIS"
	  "TRANSFER"

	  Other sections which are ignored are:

	  "DC ANALYSIS"
	  "POWER"

	2) The first line of the file contains one of the section headers OR
		some line with the word CAZM in it.

	3) All the signal names will be on one line immediately following
	   the section header.
*/

/* YCT - partial support for HSPICE "print" output */

int
load_analog(simfile, file_ptr)
	FILE *simfile;
	comm_file_t *file_ptr;
{
	run_t *curr_run, *previous;
	int one_read_legally;
	char buffer[100];
	int i, runnumber;
	double *xpts;
	
	/* for every valid segment header in this file, allocate a "run"
	   structure and fill it.  If something goes wrong while trying to
	   load this segment, trash the contents, and keep going. */

	curr_run = file_ptr->runs = (run_t *)malloc(sizeof(run_t));
	curr_run->num_X = curr_run->num_Y = 0;
	curr_run->Ynames = NULL;
	curr_run->data.analog = NULL;
	previous = NULL;

	runnumber = 0;
	one_read_legally = 0;
	while (curr_run->run_type = find_header(simfile,file_ptr->type,
		file_ptr->name)) {
		sig_file_msg(0, "	Run found and loading...","");

		if ((curr_run->num_Y=read_signames(simfile,file_ptr->type,curr_run)) &&
		 	(curr_run->num_X=read_voltages(simfile, curr_run))) {
			
			sprintf(buffer,"		Number of datapoints loaded is %d",
				curr_run->num_X);
			sig_file_msg(0, buffer,"");
			sprintf(buffer,"		Number of vectors loaded is %d",
				curr_run->num_Y);
			sig_file_msg(1, buffer,"");

			if ((curr_run->run_type == TRANSFER) &&
				(! monotonic(curr_run->data.analog[0], curr_run->num_X))) {
				curr_run->run_type = CHAR_CURVES;
				i = 1;
				xpts = curr_run->data.analog[0];
				while ((i < curr_run->num_X) && (xpts[0] != xpts[i]))
					i++;
				curr_run->num_sets = curr_run->num_X / i;
			}
			else curr_run->num_sets = 1;

			curr_run->layer = -1;
			assign_run_layer(curr_run);
			assign_runname(curr_run, file_ptr->name, runnumber);
			if (file_ptr->type == SPICE)
				decapitalize(curr_run->Ynames, curr_run->num_Y);
			synchronize_time(curr_run->Ynames[0]);
			runnumber++;

			/* alert any GP mode windows to update themselves now that we
			   have successfully loaded a new run */
			gp_update_xscales(curr_run);

			/* now allocate a new run structure to hold data from the next
			   segment in the file */
			previous = curr_run;
			curr_run->next = (run_t *)malloc(sizeof(run_t));
			curr_run = curr_run->next;
			curr_run->num_X = curr_run->num_Y = 0;
			curr_run->Ynames = NULL;
			curr_run->data.analog = NULL;
			one_read_legally = 1;
		}
	}

	if (previous != NULL)
		previous->next = NULL;
	else file_ptr->runs = NULL;
	deallocate_guts2(curr_run);
	free((char *)curr_run);

	return(one_read_legally);
}

/* Since we seem to run into a raft of incessant problems with "TIME",
   and "Time" and "time" etc., let's just standardize this vector name
   to "Time" no matter what.  Note that we assume that if it is present
   in the Ynames list, then it is the first item. */
synchronize_time(firstname)
	char *firstname;
{
	if (strlen(firstname) == 4) {
		if (((firstname[0] == 't') || (firstname[0] == 'T')) &&
		    ((firstname[1] == 'i') || (firstname[1] == 'I')) &&
		    ((firstname[2] == 'm') || (firstname[2] == 'M')) &&
		    ((firstname[3] == 'e') || (firstname[3] == 'E')))
			strcpy(firstname, "Time");
	}
}

decapitalize(Ynames, num_Y)
	char **Ynames;
	int num_Y;
{
	int i,j;

	for (i = 0; i < num_Y; i++) 
		for (j = 0; j < strlen(Ynames[i]); j++)
			Ynames[i][j] = LOWERCASE(Ynames[i][j]);
}

/* finds a header line, if any, and returns code of its type. Leaves
   simfile pointer at or above the line where signal names are 
   expected to be. */

static int
find_header(simfile, type, filename)
	FILE *simfile;
	int type;
	char *filename;
{
	char line[BUFSIZE];
	int len;
	long previous_line, ftell();

	if (type == CAZM || type == HSPICE) {
		previous_line = ftell(simfile);

		while (1) {
			if (fgets(line, BUFSIZE, simfile) == NULL) 
				return(0);
			else if ((strncmp(line, "Time", 4) == 0) ||
					 (strncmp(line, "TIME", 4) == 0) ||
					 (strncmp(line, "time", 4) == 0)) {
				/* back up so we get this again*/
				fseek(simfile, previous_line, 0);
				return(TRANSIENT);
			}
			else if(strncmp(line, "       time", 11) ==0) {
				fseek(simfile, previous_line, 0);
				return(TRANSIENT);	
			}
			else if (strncmp(line, "TRANSFER", 8) == 0)
				return(TRANSFER);
			else if (strncmp(line, "AC ANALYSIS", 11) == 0)
				return(AC);
			else if (substr(filename, ".cta", 0, strlen(filename)) != -1) {
				/* back up so we get this again*/
				fseek(simfile, previous_line, 0);
				return(TRANSFER);
			}
			else if (substr(filename, ".post", 0, strlen(filename)) != -1) {
				/* back up so we get this again*/
				fseek(simfile, previous_line, 0);
				return(TRANSFER);
			}
			previous_line = ftell(simfile);
		}
	}	
	else {	/* type is SPICE */
		while (1) {
			if (fgets(line, BUFSIZE, simfile) == NULL) 
				return(0);
			if ((strncmp(line, "0****", 5) == 0) ||
				(strncmp(line, "****", 4) == 0)) {
				len = strlen(line);
				if (substr(line, "DC TRANSFER CURVES", 0, len) != -1)
					return(TRANSFER);
				else if (substr(line, "TRANSIENT ANALYSIS", 0, len) != -1)
					return(TRANSIENT);
				else if (substr(line, "AC ANALYSIS", 0, len) != -1)
					return(AC);
			}
		}
	}
}

static int
read_signames(simfile, type, run_ptr)
	FILE *simfile;
	run_t *run_ptr;
	int type;
{
	char line[BUFSIZE];
	int i, beginning;
	int maxlen, len, sigct, lost_in_space;
	int sigptr, charptr;
	long previous_line, ftell();

	/* advance pointer to next line where, hopefully, data starts */
	if (fgets(line, BUFSIZE, simfile) == NULL) 
		return(0);

	if (type == SPICE) {
		/* scan forward in file until we find a line of stars.  The next
		   non-blank line should contain the signal names. */
		line[0] = line[1] = line[2] = '\0';
		while ((line[1] != '*') && (line[2] != '*')) 
			if (fgets(line, BUFSIZE, simfile) == NULL)
				return(0);

		/* read past all blank lines to get to the signal name line */
		if (fgets(line, BUFSIZE, simfile) == NULL)
			return(0);
		while (line[0] == '\n') {
			if (fgets(line, BUFSIZE, simfile) == NULL)
				return(0);
		}
		i=0; /* skip space before 1st signame */
		while ((isspace(line[i])) && (line[i] != '\n')) i++;
	}
	else if (type == HSPICE) {
	fgets(line, BUFSIZE, simfile);
	line[0] = 'T'; line[1] = 'i'; line[2] = 'm'; line[3] = 'e';
		i=0; 
		while ((isspace(line[i])) && (line[i] != '\n')) i++;		
	}
	else { /* for cazm file, skip blank lines  and space before first name */
		while (line[0] == '\n') { 
			if (fgets(line, BUFSIZE, simfile) == NULL)
				return(0);
		}
		i=0; /* skip space before 1st signame */
		while ((isspace(line[i])) && (line[i] != '\n')) i++;
	}

	/* now pick the names out of this line.  First, let's count how many
	   there are and how long they are.  */

	beginning = i;
	maxlen = len = sigct = 0;
	lost_in_space = 1;
	while (line[i] != '\n') {
		if (lost_in_space) {
			if (!isspace(line[i])) {
				sigct++;
				len = 1;
				lost_in_space = 0;
			}
		}
		else if (!isspace(line[i]))
				len++;
			else {
				lost_in_space = 1;
				if (len > maxlen)
					maxlen = len;
			}
		i++;
	}
	if (sigct == 0)
		return(0);
	if (len > maxlen) /* this will happen if the last name is the longest */
		maxlen = len;
	if (maxlen == 0) /* this will happen if we only read one signm */
		maxlen = len;

	run_ptr->Ynames = (char **) calloc(sigct, sizeof(char *));

	/* Now actually put the names into storage. */
	sigptr = 0; i=beginning; charptr = 0;
	while (sigptr < sigct) {
		run_ptr->Ynames[sigptr] = (char *) calloc(maxlen+1, sizeof(char));
		while (isspace(line[i])) i++;
		while (!isspace(line[i])) 
			run_ptr->Ynames[sigptr][charptr++] = line[i++];
		run_ptr->Ynames[sigptr][charptr] = '\0';
		charptr = 0;
		sigptr++;
	}
	return(sigct);
}

static int
read_voltages(simfile, run_ptr)
	FILE *simfile;
	run_t *run_ptr;
{
	double **data;
	double newv;
	int returnval, numy;
	int i, j, curr_arraysz;
	int xindex;

	while ((returnval = fscanf(simfile, "%lf", &newv)) == 0)
		fscanf(simfile,"%*s");

	if (returnval == EOF)
		return(0);

	numy = run_ptr->num_Y;

	/* allocate space for the data arrays, one for each vector */
	data = (double **) calloc(numy, sizeof(double *));
	for (i=0; i < numy; i++)
		data[i] = (double *) calloc(BLOCKSIZE, sizeof(double));

	curr_arraysz = BLOCKSIZE;

	/* process each number as it comes in. */
	xindex = 0;
	while ((returnval != EOF) && (returnval != 0)) {
		if (xindex == curr_arraysz) {
			curr_arraysz = curr_arraysz + BLOCKSIZE;

			for (j=0; j < numy; j++)
				data[j] = (double *) realloc((char *)(data[j]), 
					(sizeof(double)) * curr_arraysz);
		}
		data[0][xindex] = newv;
		for (i = 1; i < numy; i++) {
			returnval = fscanf(simfile, "%lf", &newv);
			data[i][xindex] = newv;
		}
		returnval = fscanf(simfile, "%lf", &newv);
		xindex++;
	}
	run_ptr->data.analog = data;
	return(xindex);
}
