/*---------------------------------------------------------------------------
  #######      #     #########   #          #   #    ########     #    #    #
  #            #     #            #        #    #    #            #    #    #
   ###         #    #              #      #     #    #####         #  # #  #
      ###      #    #    #####     #     #      #    #             #  # #  #
         #     #    #        #      #   #       #    #              # #  # #
  ##     #     #     ##      #       # #        #    #              ##   ##
    ####       #       ######         #         #    ########        #    #

    Designed and written by
    Lisa Pickel - pickel@mcnc.org 

    Copyright 1990, 1991 by MCNC.
    This program may be freely distributed provided that the
	copyright notice and author identification are retained.
--------------------------------------------------------------------------- */
#include "sigview.h"
#include <ctype.h>

/* ASSUMED ESIM FILE FORMAT:
	Because this is a rather nasty format to read in, assumptions are made
	about the data and no error checking is done in order to preserve speed.

	1) It is assumed that there are no blank lines in the file and that all
	   lines begin in the first column, no preceding whitespace.

	2) All signals will be listed and listed in the same order for each 
	   time step.

	3) If the "clocks" lines are present in the file, they are skipped.

	4) No spaces shall occur around the equals sign: e.g. ce=0 CANNOT be
	   ce = 0.

	5) In order to identify that we do indeed have an esim file, Sigview
	   looks for the word "transistors" in the first line.  Also, per
	   request of Cary Burnette, it will also look for "step took". Either
	   string in the first line will identify this as an esim file.

	6) A sample file is as follows:

		114 transistors, 40 nodes (0 pulled up)
		step took 31 events
		clocks=1001 9
		ce=0 rw=0 i14=X
		step took 23 events
		clocks=0110 6
		ce=0 rw=0 i14=0
		.........

	7) Only one run per file.
*/

/* returns 0 if something goes wrong. */
int
load_esim(simfile, file_ptr)
	FILE *simfile;
	comm_file_t *file_ptr;
{
	run_t *run_ptr;
	int linecount;
	char buffer[100];

	file_ptr->runs = (run_t *) malloc(sizeof(run_t));
	run_ptr = file_ptr->runs;
	run_ptr->num_X = run_ptr->num_Y = 0;
	run_ptr->Ynames = NULL;
	run_ptr->data.digital.X = NULL;
	run_ptr->data.digital.Y = NULL;

	sig_file_msg(0, "	Run found and loading...","");

	if (!(run_ptr->num_Y = read_signames(simfile, run_ptr, &linecount))) {
		file_ptr->runs = NULL;
		deallocate_guts2(run_ptr);
		free((char *)run_ptr);
		return(0);
	}
	if (!(run_ptr->num_X = read_vectors(simfile, run_ptr, linecount))) {
		file_ptr->runs = NULL;
		deallocate_guts2(run_ptr);
		free((char *)run_ptr);
		return(0);
	}

	sprintf(buffer,"		Number of vectors loaded is %d", run_ptr->num_X);
	sig_file_msg(0, buffer,"");
	sprintf(buffer,"		Number of signals loaded is %d", run_ptr->num_Y);
	sig_file_msg(1, buffer,"");

	run_ptr->layer = -1;
	assign_run_layer(run_ptr);
	assign_runname(run_ptr, file_ptr->name, 0);
	run_ptr->run_type = DIGITL;
	run_ptr->next = NULL;	/* assume that only one run per file */
	return(1);
}

/* returns 0 if something goes wrong.  returns N if all goes well,
   where N is the number of signames. Leaves FILEpointer pointing at line
   after "transistors" line.  Fills the Ynames part of the fileptr struc.
*/

static int
read_signames(simfile, run_ptr, linecount)
	FILE *simfile;
	run_t *run_ptr;
	int *linecount; /* how many lines does each dataset occupy? */
{
	int eof, done;
	int namect;
	char line[BUFSIZE];

	/* Skip first line.  We know it's there, else we wouldn't have gotten
	   into this routine.  */
	fgets(line, BUFSIZE, simfile);

	/* Skip any lines which begin "step took" OR "clocks".  Upon finding
	   a line which doesn't begin this way, extract the names from it.  */
	eof = 0; done = 0;
	while (!eof && !done) {
		if (fgets(line, BUFSIZE, simfile) == NULL) 
			eof = 1;
		else if ((substr(line,"step took", 0, 10) == -1) &&
				 (substr(line,"clocks=", 0, 8) == -1))
			done = 1;
	}
	if (eof) return(0);

	done = 0;
	namect = 0;
	*linecount = 0;
	while (!eof && !done) {
		/* for each line of names, add them to file_ptr->Ynames */
		if (extract_names(line, run_ptr, &namect))
			(*linecount)++;
		if (fgets(line, BUFSIZE, simfile) == NULL) 
			eof = 1;
		else if (substr(line,"step took", 0, 10) != -1)
			done = 1;
		else if (substr(line,"clocks=", 0, 8) != -1)
			done = 1;
	}

	rewind(simfile);
	/* Skip first line again -(the "transistors" line) if, indeed that is
	   the first line. */
	fgets(line, BUFSIZE, simfile);
	if (substr(line, "transistors", 0, strlen(line)) == -1)
		rewind(simfile);

	return(namect);
}

/* extract names from a single line from the simfile.  Change namect
   as a result. Add the names into file_ptr->Ynames.  Returns 0 if no
   names read. */

static int
extract_names(line, run_ptr, orignamect)
	char line[BUFSIZE];
	run_t *run_ptr;
	int *orignamect;
{
	int lineptr, len;
	int newnamect;
	int i,j;
	int done, afterequals;

	/* count the number of equal signs in this line.  This will tell
	   us how many signal names we have so that we can allocate the
	   right amount of space. It will also tell us the length of the
	   longest signal name in the line. */
	count_equal_signs(line, &newnamect, &len);
	if (newnamect == 0)
		return(0);

	newnamect = *orignamect + newnamect;
	if (run_ptr->Ynames == NULL)
		run_ptr->Ynames = (char **) calloc(newnamect, sizeof(char *));
	else run_ptr->Ynames = (char **) realloc((char *)run_ptr->Ynames,
			newnamect * sizeof(char *));
	if (run_ptr->Ynames == NULL) {
		sig_msg("Extract_names:  could not allocate space for sig names",
			"","");
		return(0);
	}
	len = len + 1; /* allow space for the '\0' char */
	for(i = *orignamect; i < newnamect; i++) 
		run_ptr->Ynames[i] = (char *)calloc(len, sizeof(char));

	/* now scan across the line, filling in Ynames. */
	lineptr = 0;
	i = *orignamect;
	j = 0;
	done = 0; afterequals = 0;
	while (!done) {
		switch(line[lineptr]) {
			case '\n':
			case '\0':
				done = 1;
				break;
			case ' ':
			case '\t':
				afterequals = 0;
				break;
			case '=':
				afterequals = 1;
				run_ptr->Ynames[i][j] = '\0';
				j=0;
				i++;
				break;
			default:
				if (!afterequals)
					run_ptr->Ynames[i][j++] = line[lineptr];
		}
		lineptr++;
	}

	*orignamect = newnamect;
	return(1);
}

static
count_equal_signs(line, newnamect, maxlen)
	char line[];
	int *newnamect, *maxlen;
{
	int i, len;
	int done, afterequals;

	*maxlen = 0; len = 0;
	*newnamect = 0;
	i = 0;
	done = 0; afterequals = 0;
	while (!done) {
		switch(line[i]) {
			case '\n':
			case '\0':
				done = 1;
				break;
			case ' ':
			case '\t':
				i++;
				len = 0;
				afterequals = 0;
				break;
			case '=':
				(*newnamect)++;
				if (len > *maxlen)
					*maxlen = len;
				i++;
				afterequals = 1;
				break;
			default:
				if (!afterequals)
					len++;
				i++;
		}
	}
}

/* this assumes that the simfile pointer is on the line after the 
   "transistors" line.
*/
static int
read_vectors(simfile, run_ptr, linecount)
	FILE *simfile;
	run_t *run_ptr;
	int linecount; /* how many lines does each dataset occupy? */
{
	char line[BUFSIZE];
	int i, j, xptr;
	double *X;
	char **Y;
	dig_run_t *data;
	int curr_arraysz;
	int numy, len;
	int steplen;
	double time;
	int done, yptr, afterequals;

	data = &(run_ptr->data.digital);
	numy = run_ptr->num_Y;
	if (fgets(line, BUFSIZE, simfile) == NULL)
		return(0);

	/* allocate space for the Y arrays, one for each signal */
	data->Y = (char **) calloc(numy, sizeof(char *));
	Y = data->Y;
	for (i=0; i < numy; i++)
		data->Y[i] = (char *) calloc(BLOCKSIZE, sizeof(char));

	/* allocate space for the X array */
	data->X = (double *) calloc(BLOCKSIZE, sizeof(double));
	X = data->X;
	curr_arraysz = BLOCKSIZE;

	done = 0;
	xptr = -1;
	time = 0.0;
	while (!done) {
		if (line[0] == '\n') { /* skipp this blank line */
			if (fgets(line, BUFSIZE, simfile) == NULL)
				done = 1;
		}
		else if (sscanf(line, "step took %d", &steplen) == 1) {
			xptr++;
			if (xptr == curr_arraysz) {
				curr_arraysz = curr_arraysz + BLOCKSIZE;
				data->X = (double *) realloc((char *)data->X,
					(sizeof(double)) * curr_arraysz);
				X = data->X;
				for (i=0; i < numy; i++) 
					data->Y[i] = (char *) realloc((char *)data->Y[i], 
						(sizeof(char)) * curr_arraysz);
				Y = data->Y;
			}
			X[xptr] = time;
			time = time + (double) steplen;
			if (fgets(line, BUFSIZE, simfile) == NULL)
				done = 1;
		}
		else if (substr(line, "clocks=", 0, 8) == -1) {
			/* This must be the start of a data line set. NOTE THAT WE ASSUME
			   that the format is correct, i.e. that since the first line
			   of signal data took 'linecount' lines, then all the following
			   groups of signal data take 'linecount' lines.  For the sake
			   of speed, no error checking is done! */
			yptr = 0;
			j = 0;
			while (j < linecount) {
				len = strlen(line);
				afterequals = 0;
				for(i = 0; i < len; i++) {
					switch(line[i]) {
						case '\n':
						case '\0':
							break;
						case ' ':
						case '\t':
							afterequals = 0;
							yptr++;
							break;
						case '=':
							afterequals = 1;
							break;
						default:
							if (afterequals)
								Y[yptr][xptr] = LOWERCASE(line[i]);
					}
				}
				if (fgets(line, BUFSIZE, simfile) == NULL) {
					done = 1;
					j = linecount;
				}
				j++;
			}
		}
		/* we've got a line starting with "clocks".  Skip it. */
		else if (fgets(line, BUFSIZE, simfile) == NULL) 
			done = 1;
	}
	return(xptr+1);
}
