/*---------------------------------------------------------------------------
  #######      #     #########   #          #   #    ########     #    #    #
  #            #     #            #        #    #    #            #    #    #
   ###         #    #              #      #     #    #####         #  # #  #
      ###      #    #    #####     #     #      #    #             #  # #  #
         #     #    #        #      #   #       #    #              # #  # #
  ##     #     #     ##      #       # #        #    #              ##   ##
    ####       #       ######         #         #    ########        #    #

    Designed and written by
    Lisa Pickel - pickel@mcnc.org 

    Copyright 1990, 1991 by MCNC.
    This program may be freely distributed provided that the
	copyright notice and author identification are retained.
--------------------------------------------------------------------------- */
#include "sigview.h"
#include <ctype.h>

/* ASSUMED HILO FILE FORMAT:
	1) the name "HILO " is somewhere in the text of the file.  "which_type"
	   in commonfile.c uses this fact.  NOTE that there is a blank expected
	   after "HILO ".

	2) The words "LOADED OK" appear on the line above the signal names.
	   There is no blank line between "LOADED OK" and the  names.
	   
	3) The signal names are on their sides.  Yucky, but that's the way it is.

	4) The line with "--TIME--" is separated by a blank line from the
	   last signal name line.
		
	5) A blank line always follows the "--TIME--" line, then, the data starts.

	6) Print-on-change format is fine.

	7) Either finding end of file OR the line "Total events..."  will end 
	   the data read-in.  If there's trash after the data, then "Total
	   events..." had better be there!!
*/

int
load_hilo(simfile, file_ptr)
	FILE *simfile;
	comm_file_t *file_ptr;
{
	run_t *run_ptr;
	char buffer[100];

	file_ptr->runs = (run_t *)malloc(sizeof(run_t));
	run_ptr = file_ptr->runs;
	run_ptr->num_X = run_ptr->num_Y = 0;
	run_ptr->Ynames = NULL;
	run_ptr->data.digital.X = NULL;
	run_ptr->data.digital.Y = NULL;
	sig_file_msg(0, "	Run found and loading...","");

	if (!(run_ptr->num_Y = read_signames(simfile, run_ptr))) {
		file_ptr->runs = NULL;
		deallocate_guts2(run_ptr);
		free((char *)run_ptr);
		return(0);
	}
	if (!(run_ptr->num_X = read_vectors(simfile, run_ptr))) {
		file_ptr->runs = NULL;
		deallocate_guts2(run_ptr);
		free((char *)run_ptr);
		return(0);
	}

	sprintf(buffer,"		Number of vectors loaded is %d", run_ptr->num_X);
	sig_file_msg(0, buffer,"");
	sprintf(buffer,"		Number of signals loaded is %d", run_ptr->num_Y);
	sig_file_msg(1, buffer,"");

	run_ptr->layer = -1;
	assign_run_layer(run_ptr);
	assign_runname(run_ptr, file_ptr->name, 0);
	run_ptr->run_type = DIGITL;
	run_ptr->next = NULL;	/* assume that only one run per file */
	return(1);
}

/* returns 0 if something goes wrong.  returns N if all goes well,
   where N is the number of signames.  Fills run_ptr->Ynames with the
   signal names found in the file.  Positions the file pointer at the
   start of data and leaves it there for read-vectors to pick up.
*/


static int
read_signames(simfile, run_ptr)
	FILE *simfile;
	run_t *run_ptr;
{
	int eof, done;
	int row, column;
	int i, j, namect;
	char line[BUFSIZE];
	int namesize;
	long namebegin, ftell(), nameend;
	char **lines;
	int namepos[BUFSIZE];

	/*	Skip to line which has "LOADED OK" in it */
	eof = 0; done = 0;
	while (!eof && !done) {
		if (fgets(line, BUFSIZE, simfile) == NULL) 
			eof = 1;
		else if (substr(line,"LOADED OK",0,strlen(line)) != -1) 
			done = 1;
	}
	if (eof) return(0);

	namebegin = ftell(simfile);

	/* Now get next lines which should have parts of the signames in them.
	   Count how many of these lines there are.  This is the largest size any 
	   signal name could have. */

	eof = 0; done = 0;
	namesize = 0;
	while (!eof && !done) {
		if (fgets(line, BUFSIZE, simfile) == NULL) 
			eof = 1;
		else if (substr(line,"--TIME--",0, strlen(line)) != -1) 
			done = 1;
		else namesize++;
	}
	namesize--;	/* subtract one for the blank line above "--TIME--" */

	nameend = ftell(simfile);
	/* now read and store these lines containing the signal names */

	fseek(simfile, namebegin, 0);
	lines = (char **)calloc(namesize, sizeof(char *));
	for (i=0; i < namesize; i++) { 
		lines[i] = (char *)calloc(BUFSIZE, sizeof(char));
		fgets(lines[i], BUFSIZE, simfile);
	}

	/* count how many names we have by looking at the last line of names
	   we read in.  This last line will contain a character for every
	   signal. */
	column = 0; namect = 0;
	while ((column < BUFSIZE) && (lines[namesize-1][column] != '\n')) {
		if (!isspace(lines[namesize-1][column])) {
			namepos[namect] = column;
			namect++;
		}
		column++;
	}
	run_ptr->Ynames = (char **) calloc(namect, sizeof(char *));

	/* Now actually put the names into storage. */
	i=0;
	while (i < namect) {
		run_ptr->Ynames[i] = (char *) calloc(namesize+1, sizeof(char));
		j = 0;
		for (row = 0; row < namesize; row++) {
			if (!isspace(lines[row][namepos[i]])) 
				run_ptr->Ynames[i][j++] = lines[row][namepos[i]];
		}
		run_ptr->Ynames[i][j] = '\0';
		i++;
	}

	/* Now skip to the "--TIME--" line and then the  blank line which we assume 
	   follows the "--TIME--" line.  This is so we are left in position to 
	   read the numbers in the file. */
	fseek(simfile, nameend, 0);
	if (fgets(lines[0], BUFSIZE, simfile) == NULL) 
		return(0);	

	return(namect);
}

/* places the actual 0s and 1s into the run_ptr structure.  */
static int
read_vectors(simfile, run_ptr)
	FILE *simfile;
	run_t *run_ptr;
{
	int eof, get_line();
	int i, numy, vec_ct;
	double newtime;
	char newvals[BUFSIZE];
	double *X;
	char **Y;
	dig_run_t *data;
	int curr_arraysz;

	data = &(run_ptr->data.digital);
	numy = run_ptr->num_Y;

	/* allocate space for the Y arrays, one for each signal */
	data->Y = (char **) calloc(numy, sizeof(char *));
	Y = data->Y;
	for (i=0; i < numy; i++)
		data->Y[i] = (char *) calloc(BLOCKSIZE, sizeof(char));

	/* allocate space for the X array */
	data->X = (double *) calloc(BLOCKSIZE, sizeof(double));
	X = data->X;
	
	curr_arraysz = BLOCKSIZE;

	eof = get_line(simfile, numy, &newtime, newvals);
	vec_ct = 0;
	while(!eof) {
		if (vec_ct == curr_arraysz) {

			curr_arraysz = curr_arraysz + BLOCKSIZE;

			data->X = (double *)realloc((char *)data->X,
				(sizeof(double)) * curr_arraysz);
			for (i=0; i < numy; i++)
				data->Y[i] = (char *) realloc((char *)(data->Y[i]), 
					(sizeof(char)) * curr_arraysz);
			X = data->X;
			Y = data->Y;
		}

		X[vec_ct] = newtime;
		for (i=0; i < numy; i++) 
			Y[i][vec_ct] = newvals[i];
		
		vec_ct++;
		eof = get_line(simfile, numy, &newtime, newvals);
	}

	return(vec_ct);
}

/* return 1 if we hit end of file or some other problem occured.  Else
   return 0.  This routine expects that we have positioned the file pointer
   at the data.  */

static int
get_line(simfile, num_y, time, newvals)
	FILE *simfile;
	int num_y;
	double *time;
	char newvals[];
{
	int i;
	int rc;
	char c[2];

	rc = fscanf(simfile, "%lf", time);

	if ((rc == EOF) || (rc == 0)) 
		return(1);

	for (i = 0; i < num_y; i++) {
		if (fscanf(simfile, "%1s", c) == EOF) 
			return(1);
		else newvals[i] = LOWERCASE(c[0]);
	}
	return(0);
}
