/*---------------------------------------------------------------------------
  #######      #     #########   #          #   #    ########     #    #    #
  #            #     #            #        #    #    #            #    #    #
   ###         #    #              #      #     #    #####         #  # #  #
      ###      #    #    #####     #     #      #    #             #  # #  #
         #     #    #        #      #   #       #    #              # #  # #
  ##     #     #     ##      #       # #        #    #              ##   ##
    ####       #       ######         #         #    ########        #    #

    Designed and written by
    Lisa Pickel - pickel@mcnc.org 

    Copyright 1990, 1991 by MCNC.
    This program may be freely distributed provided that the
    copyright notice and author identification are retained.
--------------------------------------------------------------------------- */

#include "sigview.h"
#include <ctype.h>

/* ASSUMED RAW FILE FORMAT:

    Both Spice and CAZM produce this same file format.  It looks as
    follows:

    Title
    Date
    Name
    Flags
    No. Variables: 3
    No. Points: 00061
    Command
    Variables:
        0  time time
        1  v(2) voltage
        2  v(5) voltage
    Values:
    0   0e0
        0e0
        0e0
    1   5e-10
        0e0
        0e0
    2   1e-9
        0e0
        4.7e-1

    For now, the only lines I pay attention to are "No. Variables",
    "No. Points", "Variables", and "Values".  I scan for those exact
    words and then assume that the information to follow is in the
    format above.

    Multiple runs may be present in a file.

    If Variable 0 is time, then I assume this is a TRANSIENT run,
    else if Variable 0 is Frequency, then I assume this is an AC
    run, else I assume it is a TRANSFER run.
*/

/* JTS - added support for Spectre "nutascii" format */

int
load_raw(simfile, file_ptr)
    FILE *simfile;
    comm_file_t *file_ptr;
{
    run_t *curr_run, *previous;
    int i, runnumber, one_read_legally;
    char buffer[100];
    double *xpts;

    /* for every valid segment header in this file, allocate a "run"
       structure and fill it.  If something goes wrong while trying to
       load this segment, trash the contents, and keep going. */

    curr_run = file_ptr->runs = (run_t *)malloc(sizeof(run_t));
    curr_run->num_X = curr_run->num_Y = 0;
    curr_run->Ynames = NULL;
    curr_run->data.analog = NULL;
    previous = NULL;

    runnumber = 0;
    one_read_legally = 0;
    while (find_header(simfile)) {
        sig_file_msg(0, "   Run found and loading...","");

        if (read_stats(simfile, curr_run) &&
            read_vecnames(simfile,curr_run) &&
            read_values(simfile, curr_run)) {
            
            sprintf(buffer,"        Number of datapoints loaded is %d",
                curr_run->num_X);
            sig_file_msg(0, buffer,"");
            sprintf(buffer,"        Number of vectors loaded is %d",
                curr_run->num_Y);
            sig_file_msg(1, buffer,"");
            curr_run->layer = -1;
            assign_run_layer(curr_run);
            assign_runname(curr_run, file_ptr->name, runnumber);
            if ((curr_run->run_type == AC) && (file_ptr->type == SPICE))
                decapitalize(curr_run->Ynames, curr_run->num_Y);
            runnumber++;

            if ((curr_run->run_type == TRANSFER) &&
                (! monotonic(curr_run->data.analog[0], curr_run->num_X))) {
                curr_run->run_type = CHAR_CURVES;
                i = 1;
                xpts = curr_run->data.analog[0];
                while ((i < curr_run->num_X) && (xpts[0] != xpts[i]))
                    i++;
                curr_run->num_sets = curr_run->num_X / i;
            }
            else curr_run->num_sets = 1;

            /* alert any GP mode windows to update themselves now that we
               have successfully loaded a new run */
            gp_update_xscales(curr_run);

            /* now allocate a new run structure to hold data from the next
               segment in the file */
            previous = curr_run;
            curr_run->next = (run_t *)malloc(sizeof(run_t));
            curr_run = curr_run->next;
            curr_run->num_X = curr_run->num_Y = 0;
            curr_run->Ynames = NULL;
            curr_run->data.analog = NULL;
            one_read_legally = 1;
        }
    }

    if (previous != NULL)
        previous->next = NULL;
    else file_ptr->runs = NULL;
    deallocate_guts2(curr_run);
    free((char *)curr_run);

    return(one_read_legally);
}

/* finds the line "Title:....".  Leaves simfile pointer above the line 
   "No. Variables". */

static int
find_header(simfile)
    FILE *simfile;
{
    char line[BUFSIZE];

    /* find the line beginning "Title:" */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);
    while (strncmp(line, "Title:", 6) != 0) {
        if (fgets(line, BUFSIZE, simfile) == NULL) 
            return(0);
    }

    /* Now realign at the line before No. Variables, which is Flags */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);
    while(strncmp(line, "Flags:", 6) != 0) {
        if (fgets(line, BUFSIZE, simfile) == NULL) 
            return(0);
    }
    return(1);
}

/* return 0 if we don't successfully read No. Variables and No. Points
   lines.  Leave file pointer at the "Variables" line. */
static int
read_stats(simfile, run_ptr)
    FILE *simfile;
    run_t *run_ptr;
{
    char line[BUFSIZE];
    int sigct, pointct;
    char no[4], dummy[10];
    long filepos; /* JTS */

    /* advance pointer to next line where, hopefully, No. Variables is
       located */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);
    if (sscanf(line, "%s %s %d", no, dummy, &sigct) != 3)
        return(0);
    if (sigct <= 0)
        return(0);
    run_ptr->num_Y = sigct;

    /* now advance to No. Points */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);
    if (sscanf(line, "%s %s %d", no, dummy, &pointct) != 3)
        return(0);
    if (pointct <= 0)
        return(0);
    run_ptr->num_X = pointct;

    /* JTS */
    filepos = ftell( simfile );

    /* skip the "Command" line */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);

    /* JTS */
    if( strncmp( line, "Variables:", 10 ) == 0 ) {
        /* this might be a Spectre file -> no "Command" line */
        fseek( simfile, filepos, SEEK_SET );
    }

    return(1);
}

static int
read_vecnames(simfile, run_ptr)
    FILE *simfile;
    run_t *run_ptr;
{
    char line[BUFSIZE];
    int num, i;
    char vecnm[SIGNMSIZE], vectype[10];

    /* read the "Variables" line */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);
    if (strncmp(line, "Variables:", 10) != 0)
        return(0);

    run_ptr->Ynames = (char **) calloc(run_ptr->num_Y, sizeof(char *));

    /* Now actually put the names into storage. */
    for (i = 0; i < run_ptr->num_Y; i++) {
        /* JTS */
        if (i != 0 || 
            sscanf(line+10, "%d %s %s", &num,vecnm,vectype) != 3) {
            if (fgets(line, BUFSIZE, simfile) == NULL) 
                return(0);
            if (sscanf(line, "%d %s %s", &num, vecnm, vectype) != 3)
                return(0);
        }
        run_ptr->Ynames[i] = (char *) calloc(strlen(vecnm)+1, sizeof(char));
        strcpy(run_ptr->Ynames[i], vecnm);

        /* If this is the first variable, then determine type of run */
        if (i == 0) {
            if ((strcmp(vectype, "Frequency") == 0) ||
                (strcmp(vectype, "frequency") == 0))
                run_ptr->run_type = AC;
            else if ((strcmp(vectype, "time") == 0) ||
                      (strcmp(vectype, "s") == 0))   /* JTS */
                run_ptr->run_type = TRANSIENT;
            else run_ptr->run_type = TRANSFER;
        }
    }
    return(1);
}

static int
read_values(simfile, run_ptr)
    FILE *simfile;
    run_t *run_ptr;
{
    char line[BUFSIZE];
    int numy, numx;
    double **data;
    int i, j;
    int xindex;
    int mistake;

    /* read the "Values" line */
    if (fgets(line, BUFSIZE, simfile) == NULL) 
        return(0);
    if (strncmp(line, "Values:", 7) != 0)
        return(0);

    numy = run_ptr->num_Y;
    numx = run_ptr->num_X;

    /* allocate space for the data arrays, one for each vector */
    data = (double **) calloc(numy, sizeof(double *));
    for (i = 0; i < numy; i++)
        data[i] = (double *) calloc(numx, sizeof(double));

    mistake = 0;
    i = 0;
    while ((i < numx) && (! mistake)) {
        if (fscanf(simfile, "%d", &xindex) != 1)
            mistake = 1;

        /* now read the values for this index */
        j = 0;
        while ((j < numy) && (!mistake)) {
            if (fscanf(simfile, "%lf", &(data[j][i])) != 1)
                mistake = 1;
            j++;
        }
        i++;
    }

    if (mistake) {
        for (i = 0; i < numy; i++)
            free ((char *) data[i]);
        free((char *) data);
        return(0);
    }

    run_ptr->data.analog = data;
    return(1);
}
