/*---------------------------------------------------------------------------
  #######      #     #########   #          #   #    ########     #    #    #
  #            #     #            #        #    #    #            #    #    #
   ###         #    #              #      #     #    #####         #  # #  #
      ###      #    #    #####     #     #      #    #             #  # #  #
         #     #    #        #      #   #       #    #              # #  # #
  ##     #     #     ##      #       # #        #    #              ##   ##
    ####       #       ######         #         #    ########        #    #

    Designed and written by
    Lisa Pickel - pickel@mcnc.org 

    Copyright 1990, 1991 by MCNC.
    This program may be freely distributed provided that the
	copyright notice and author identification are retained.
--------------------------------------------------------------------------- */
#include "sigview.h"

/* ASSUMED RNL/LDV FILE FORMAT:

   1) The first line is a "marker": 

	  >>> somenumber somename  
	  
	  I'm not sure, but it looks like there could be others.  In any case,
	  I skip all of these marker lines when they are encountered.  NOTE 
	  that this implies that multiple runs within the same file ARE NOT
	  ALLOWED.

   2) The next line begins the data.  It is expected that the user had
	  done a "(chflag nodes)" in order to insure print on change into this
	  file.  The data looks like:
		nodename     value (either 0,1,or x)     timestep

	  Note that timestep is read in as a double to insure that numbers like
	  "6.22" can be handled.  Further it is assumed that these timesteps are
	  always increasing.  That is, you canNOT have:
		a1 x 4
		a2 x 4.7
		a3 x 4.2  *** less than 4.7.  This is a no no 

   3) Data is read until end of file is reached.

   4) It is expected that all signals will be mentioned for the first
	  timestep (which is usually zero).  I.e. for a file which contains 
	  signals "a b out bert", the first four lines of data MUST give the 
	  values of all these signals at the first time (order not important):
			b 0 0
			a 1 0
			out x 0
			bert x 0
*/

int
load_rnl(simfile, file_ptr)
	FILE *simfile;
	comm_file_t *file_ptr;
{
	run_t *run_ptr;
	char buffer[100];

	file_ptr->runs = (run_t *) malloc(sizeof(run_t));
	run_ptr = file_ptr->runs;
	run_ptr->num_X = run_ptr->num_Y = 0;
	run_ptr->Ynames = NULL;
	run_ptr->data.digital.X = NULL;
	run_ptr->data.digital.Y = NULL;
	sig_file_msg(0, "	Run found and loading...","");

	if (!(run_ptr->num_Y = read_signames(simfile, run_ptr))) {
		file_ptr->runs = NULL;
		deallocate_guts2(run_ptr);
		free((char *)run_ptr);
		return(0);
	}
	if (!(run_ptr->num_X = read_vectors(simfile, run_ptr))) {
		file_ptr->runs = NULL;
		deallocate_guts2(run_ptr);
		free((char *)run_ptr);
		return(0);
	}

	sprintf(buffer,"		Number of vectors loaded is %d", run_ptr->num_X);
	sig_file_msg(0, buffer,"");
	sprintf(buffer,"		Number of signals loaded is %d", run_ptr->num_Y);
	sig_file_msg(1, buffer,"");

	run_ptr->layer = -1;
	assign_run_layer(run_ptr);
	assign_runname(run_ptr, file_ptr->name, 0);
	run_ptr->run_type = DIGITL;
	run_ptr->next = NULL;	/* assume that only one run per file */
	return(1);
}

static int
read_signames(simfile, run_ptr)
	FILE *simfile;
	run_t *run_ptr;
{
	int sigct;
	double time;
	char line[BUFSIZE];
	char signm[SIGNMSIZE];
	int done;
	double first_time;
	char c[2];
	int allocated;

	allocated = 10;
	run_ptr->Ynames = (char **) calloc(allocated, sizeof(char *));

	/* read through file entries for the first timestep.  Collect the
	   signal names in Ynames and count how many there are. */

	sigct = 0; done = 0;
	while (!done) {
		if (fgets(line, BUFSIZE, simfile) == NULL)
			done = 1;

		if ((line[0] == '>') && (substr(line, ">>>", 0, 3) != -1))
			;
		else if (sscanf(line, "%s %s %lf", signm, c, &time) == 3) {
			if (sigct == 0)
				first_time = time;
			if (time != first_time) 
				done = 1;
			else {
				if (sigct == allocated) {
					allocated = allocated + 10;
					run_ptr->Ynames=(char **)realloc((char *)run_ptr->Ynames, 
						(sizeof(char *)) * allocated);
				}
				run_ptr->Ynames[sigct] = (char *)calloc(strlen(signm)+1,
											sizeof(char));
				strcpy(run_ptr->Ynames[sigct], signm);
				sigct++;
			}
		}
	}

	if (sigct != 0)
		run_ptr->Ynames = (char **) realloc((char *)run_ptr->Ynames,
			sigct * sizeof(char *));

	rewind(simfile);
	return(sigct);
}

static int
read_vectors(simfile, run_ptr)
	FILE *simfile;
	run_t *run_ptr;
{
	char line[BUFSIZE];
	char signm[SIGNMSIZE];
	char c[2];
	double time;
	int x_ptr, allocated;
	int numy;
	dig_run_t *data;
	double *X;
	int i,j;

	data = &(run_ptr->data.digital);
	numy = run_ptr->num_Y;
	
	/* allocate space for the X array. */
	x_ptr = -1;
	allocated = BLOCKSIZE;
	data->X = (double *) calloc(allocated, sizeof(double));
	X = data->X;

	/* allocate space for the Y arrays, one for each signal */
	data->Y = (char **) calloc(numy, sizeof(char *));
	for (i=0; i < numy; i++) {
		data->Y[i] = (char *) calloc(allocated, sizeof(char));
		for (j = 0; j < allocated; j++)
			data->Y[i][j] = '\0';
	}
	/* fill the X and Y arrays. */
	while (fgets(line, BUFSIZE, simfile) != NULL) {

		if ((line[0] == '>') && (substr(line, ">>>", 0, 3) != -1))
			;
		else if (sscanf(line, "%s %s %lf", signm, c, &time) == 3) {
			if (x_ptr == -1) {
				x_ptr++;
				X[0] = time;
				store_data(signm, c[0], x_ptr, run_ptr->Ynames, data->Y, numy);
			}	
			else if (time != X[x_ptr]) {
				x_ptr++;
				if (x_ptr == allocated) {
					allocated = allocated + BLOCKSIZE;
					data->X = (double *)
						realloc((char *)data->X,(sizeof(double)) * allocated);
					X = data->X;
					for (i=0; i < numy; i++) {
						data->Y[i] = (char *) realloc((char *)data->Y[i], 
							(sizeof(char)) * allocated);
/* IS THERE A REASON FOR THIS INIT?  IT IS WASTEFUL */
						for (j = allocated-BLOCKSIZE; j < allocated; j++)
							data->Y[i][j] = '\0';
					}
				}
				X[x_ptr] = time;
				store_data(signm, c[0], x_ptr, run_ptr->Ynames, data->Y, numy);
			}
			else store_data(signm, c[0], x_ptr, run_ptr->Ynames, data->Y, numy);
		}
	}
	if (x_ptr > -1)
		fill_holes(data->Y, x_ptr+1, numy);

	return(x_ptr + 1);
}


store_data(signm, c, x_ptr, Ynames, Y, numy)
	char *signm, c;
	int x_ptr, numy;
	char **Ynames;
	char **Y;
{
	int i;
	int done;

	/* find signm in Ynames to get the index into Y.  Place value into
	   corresponding Y array. */

	i = 0; done = 0;
	while (!done) {
		if (i == numy) {
			sig_msg("Load: signal name ",signm,
				" found which was not mentioned at the first timestep");
			done = 1;
		}
		else if ((Ynames[i][0] == signm[0]) && 
				 (strcmp(Ynames[i], signm) == 0)) {
			done = 1;	
			Y[i][x_ptr] = LOWERCASE(c);
		}
		else i++;
	}

}

/* we fill in holes in each Y array.  Note that we assume that for the 
   first time given in the file, there was a value listed in the file for
   every signal. */

fill_holes(Y, numx, numy)
	char **Y;
	int numx, numy;
{
	int i,j;
	char propagate_char;

	for (i = 0; i < numy; i++) { 	
		propagate_char = Y[i][0]; /* this had better be an actual value!! */
		for (j = 1; j < numx; j++) {
			if (Y[i][j] == '\0')
				Y[i][j] = propagate_char;
			else propagate_char = Y[i][j];
		}
	}
}
